extern "C" {
#include "modkandinsky.h"
#include <py/runtime.h>
}
#include <kandinsky.h>
#include "port.h"


static int strlen(const char *s_){
  const char * s=s_;
  while (*s)
    ++s;
  return s-s_;
}

static mp_obj_t TupleForKDColor(KDColor c) {
  mp_obj_tuple_t * t = static_cast<mp_obj_tuple_t *>(MP_OBJ_TO_PTR(mp_obj_new_tuple(3, NULL)));
  t->items[0] = MP_OBJ_NEW_SMALL_INT(c.red());
  t->items[1] = MP_OBJ_NEW_SMALL_INT(c.green());
  t->items[2] = MP_OBJ_NEW_SMALL_INT(c.blue());
  return MP_OBJ_FROM_PTR(t);
}

/* KDIonContext::sharedContext needs to be set to the wanted Rect before
 * calling kandinsky_get_pixel, kandinsky_set_pixel and kandinsky_draw_string.
 * We do this here with displaySandbox(), which pushes the SandboxController on
 * the stackViewController and forces the window to redraw itself.
 * KDIonContext::sharedContext is set to the frame of the last object drawn. */

mp_obj_t modkandinsky_color(size_t n_args, const mp_obj_t *args) {
  mp_obj_t color;
  if (n_args == 1) {
    color = args[0];
  } else if (n_args == 2) {
    mp_raise_TypeError("color takes 1 or 3 arguments");
    return mp_const_none;
  } else {
    assert(n_args == 3);
    color = mp_obj_new_tuple(n_args, args);
  }
  return TupleForKDColor(MicroPython::Color::Parse(color));
}

/* Calling ExecutionEnvironment::displaySandbox() hides the console and switches
 * to another mode. So it's a good idea to retrieve and handle input parameters
 * before calling displaySandbox, otherwise error messages (such as TypeError)
 * won't be visible until the user comes back to the console screen. */

mp_obj_t modkandinsky_get_pixel(mp_obj_t x, mp_obj_t y) {
  KDPoint point(mp_obj_get_int(x), mp_obj_get_int(y));
  KDColor c;
  KDIonContext::sharedContext()->getPixel(point, &c);
  return TupleForKDColor(c);
}

mp_obj_t modkandinsky_set_pixel(mp_obj_t x, mp_obj_t y, mp_obj_t input) {
  KDPoint point(mp_obj_get_int(x), mp_obj_get_int(y));
  KDColor kdColor = MicroPython::Color::Parse(input);
  MicroPython::ExecutionEnvironment::currentExecutionEnvironment()->displaySandbox();
  KDIonContext::sharedContext()->setPixel(point, kdColor);
  return mp_const_none;
}

//TODO Use good colors
mp_obj_t modkandinsky_draw_string(size_t n_args, const mp_obj_t * args) {
  const char * text = mp_obj_str_get_str(args[0]);
  KDPoint point(mp_obj_get_int(args[1]), mp_obj_get_int(args[2]));
  KDColor textColor = (n_args >= 4) ? MicroPython::Color::Parse(args[3]) : KDColorBlack;
  KDColor backgroundColor = (n_args >= 5) ? MicroPython::Color::Parse(args[4]) : KDColorWhite;
  MicroPython::ExecutionEnvironment::currentExecutionEnvironment()->displaySandbox();
  KDIonContext::sharedContext()->drawString(text, point, KDFont::LargeFont, textColor, backgroundColor);
  /* Before and after execution of "modkandinsky_draw_string",
   * "micropython_port_vm_hook_loop" is called by "mp_execute_bytecode" and will
   * call "micropython_port_interrupt_if_needed" every 20000 calls.
   * However, "drawString" function might take some time to execute leading to
   * drastically decrease the frequency of calls to
   * "micropython_port_vm_hook_loop" and thus to
   * "micropython_port_interrupt_if_needed". So we add an extra
   * check for user interruption here. This way the user can still interrupt an
   * infinite loop calling 'drawString' for instance. */
  micropython_port_interrupt_if_needed();
  return mp_const_none;
}

mp_obj_t modkandinsky_fill_rect(size_t n_args, const mp_obj_t * args) {
  mp_int_t x = mp_obj_get_int(args[0]);
  mp_int_t y = mp_obj_get_int(args[1]);
  mp_int_t width = mp_obj_get_int(args[2]);
  mp_int_t height = mp_obj_get_int(args[3]);
  if (width < 0) {
    width = -width;
    x = x - width;
  }
  if (height < 0) {
    height = -height;
    y = y - height;
  }
  KDRect rect(x, y, width, height);
  KDColor color = MicroPython::Color::Parse(args[4]);
  MicroPython::ExecutionEnvironment::currentExecutionEnvironment()->displaySandbox();
  KDIonContext::sharedContext()->fillRect(rect, color);
  // Cf comment on modkandinsky_draw_string
  micropython_port_interrupt_if_needed();
  return mp_const_none;
}


/* C-- SDK (c) B. Parisse, 2019 */

#define LCD_WIDTH_PX 320
#define LCD_HEIGHT_PX 222

void numworks_set_pixel(int x, int y, int color) {
  if (x<0 || x>=LCD_WIDTH_PX || y<0 || y>=LCD_HEIGHT_PX)
    return;
  auto ctx=KDIonContext::sharedContext();
  KDRect save=ctx->m_clippingRect;
  KDPoint o=ctx->m_origin;
  ctx->setClippingRect(KDRect(0,0,320,240));
  ctx->setOrigin(KDPoint(0,0));
  KDColor c(color);
  KDPoint point(x,y+18);
  KDIonContext::sharedContext()->pushRect(KDRect(point, 1, 1), &c);
  ctx->setClippingRect(save);
  ctx->setOrigin(o);  
}

void numworks_fill_rect(int x,int y,int w,int h,int c){
  KDColor color = c;
  auto ctx=KDIonContext::sharedContext();
  KDRect save=ctx->m_clippingRect;
  KDPoint o=ctx->m_origin;
  ctx->setClippingRect(KDRect(0,0,320,240));
  ctx->setOrigin(KDPoint(0,0));
#if 1
  if (x<0){
    w += x;
    x=0;
  }
  if (y<0){
    h += y;
    y=0;
  }
  if (h+y>=LCD_HEIGHT_PX)
    h=LCD_HEIGHT_PX-y;
  if (x+w>=LCD_WIDTH_PX)
    w=LCD_WIDTH_PX-x;
  if (h<=0 || w<=0)
    return;
  KDRect rect(x,y+18,w,h);
  KDIonContext::sharedContext()->pushRectUniform(rect,color);
#else
  KDRect rect(x,y,w,h); 
  KDIonContext::sharedContext()->fillRect(rect, color);
#endif
  ctx->setClippingRect(save);
  ctx->setOrigin(o);  
}

int numworks_get_pixel(int x, int y) {
  KDPoint point(x,y);
  KDColor c;
  KDIonContext::sharedContext()->getPixel(point,&c);
  return c;
}

int numworks_draw_string(int x,int y,int c,int bg,const char * text,bool fake){
  if (fake)
    return x+10*strlen(text);
  auto ptr=MicroPython::ExecutionEnvironment::currentExecutionEnvironment();
  KDPoint point(x,y);
  if (ptr)
    ptr->displaySandbox();
  auto ctx=KDIonContext::sharedContext();
  KDRect save=ctx->m_clippingRect;
  KDPoint o=ctx->m_origin;
  ctx->setClippingRect(KDRect(0,18,fake?0:320,fake?0:222));
  ctx->setOrigin(KDPoint(0,18));
  point=KDIonContext::sharedContext()->drawString(text, point, KDFont::LargeFont, c, bg);
  ctx->setClippingRect(save);
  ctx->setOrigin(o);
  return point.x();
}

int numworks_draw_string_small(int x,int y,int c,int bg,const char * text,bool fake){
  auto ptr=MicroPython::ExecutionEnvironment::currentExecutionEnvironment();
  KDPoint point(x,y);
  if (ptr)
    ptr->displaySandbox();
  auto ctx=KDIonContext::sharedContext();
  KDRect save=ctx->m_clippingRect;
  KDPoint o=ctx->m_origin;
  ctx->setClippingRect(KDRect(0,18,320,fake?0:222));
  ctx->setOrigin(KDPoint(0,18));
  point=ctx->drawString(text, point, KDFont::SmallFont, c, bg);
  ctx->setClippingRect(save);
  ctx->setOrigin(o);
  return point.x();
}

void numworks_hide_graph(){
  auto ptr=MicroPython::ExecutionEnvironment::currentExecutionEnvironment();
  if (ptr)
    ptr->resetSandbox();
}

void numworks_show_graph(){
  auto ptr=MicroPython::ExecutionEnvironment::currentExecutionEnvironment();
  if (ptr)
    ptr->displaySandbox();
}
